/*
 * Copyright (c) 2018, 2022, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package org.graalvm.component.installer.remote;

import java.io.IOException;
import java.net.ConnectException;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;
import org.graalvm.component.installer.Feedback;
import org.graalvm.component.installer.URLConnectionFactory;
import java.io.BufferedReader;
import java.io.InputStreamReader;

/**
 * Creates URLConnections to the given destination. Caches the decision about proxy. For the first
 * {@link #openConnection(java.net.URI, org.graalvm.component.installer.URLConnectionFactory.Configure)},
 * the code wil open
 * <ul>
 * <li>a direct connection to the proxy
 * <li>a connection through http proxy, if configured
 * <li>a connection through https proxy, if configured
 * </ul>
 * Users may forget to set both the http/https proxies, so the code will attempt to use whetaver is
 * known. All those connections will attempt to connect to the destination. The first connection
 * attempt that succeeds will be used. The mechanism (proxy setting, direct) that established the
 * connection will be cached for subsequent requests to the same authority. Additional requests to
 * the same location should not open additional probes then.
 * <p/>
 * The factory is caching threads; initially at most 3 threads will be created for probes, then each
 * request will reuse a thread from the thread pool (to watch time out the operation on the main
 * thread).
 * 
 * @author sdedic
 */
public class ProxyConnectionFactory implements URLConnectionFactory {
    /**
     * The max delay to connect to the final destination or open a proxy connection. In seconds.
     */
    private static final int DEFAULT_CONNECT_DELAY = Integer.getInteger("org.graalvm.component.installer.connectDelaySec", 10);

    /**
     * Delay in the case proxies are not used. Heuristics is not used at all, so the connection time
     * may be longer. In seconds.
     */
    private static final int DEFAULT_DIRECT_CONNECT_DELAY = Integer.getInteger("org.graalvm.component.installer.directConnectDelaySec", 20);

    private static final String PROXY_SCHEME_PREFIX = "http://"; // NOI18N

    private final Feedback feedback;
    private final URL urlBase;

    // @GuardedBy(this)
    private Connector winningConnector;

    /**
     * Thread pool for connection attempts. Subsequent connections do not send unnecessary probes,
     * but are done in a thread so the main thread may time out the operation. Most likely the
     * threads will be allocated from this pool.
     */
    private final ExecutorService connectors = Executors.newCachedThreadPool();

    /**
     * HTTP proxy settings. The default is taken from system environment variables and system
     * property
     */
    String envHttpProxy = System.getProperty("http_proxy", System.getenv("http_proxy")); // NOI18N

    /**
     * HTTPS proxy settings. The default is taken from system environment variables and system
     * property
     */
    String envHttpsProxy = System.getProperty("https_proxy", System.getenv("https_proxy")); // NOI18N

    /**
     * The configurable delay for this factory. Initialized to {@link #DEFAULT_CONNECT_DELAY}.
     */
    private int connectDelay = DEFAULT_CONNECT_DELAY;

    private int directConnectDelay = DEFAULT_DIRECT_CONNECT_DELAY;

    public ProxyConnectionFactory(Feedback feedback, URL urlBase) {
        this.feedback = feedback.withBundle(ProxyConnectionFactory.class);
        this.urlBase = urlBase;
    }

    /**
     * Customizes connection timeouts. Base delay must be at least 0 (infinite wait). If the direct
     * delay is negative, it is scaled from base delay using the same factor as the default timeouts
     * 
     * @param delay base delay, 0 for infinite wait. Must not be negative
     * @param directDelay single connection delay, negative to derive from the base delay
     */
    public void setConnectDelay(int delay, int directDelay) {
        if (delay < 0) {
            throw new IllegalArgumentException();
        }
        this.connectDelay = delay;
        if (directDelay >= 0) {
            this.directConnectDelay = directDelay;
        } else {
            // scale the base delay by the default connection factor
            float factor = Math.min(1, ((float) DEFAULT_DIRECT_CONNECT_DELAY / DEFAULT_CONNECT_DELAY));
            this.directConnectDelay = Math.round(delay * factor);
        }
    }

    /**
     * Scales the default timeouts by some factor.
     * 
     * @param factor factor to scale the default timeouts.
     */
    public void setConnectDelayFactor(float factor) {
        this.connectDelay = Math.round(DEFAULT_CONNECT_DELAY * factor);
        this.directConnectDelay = Math.round(DEFAULT_DIRECT_CONNECT_DELAY * factor);
    }

    public ProxyConnectionFactory setProxy(boolean secure, String proxyURI) {
        if (secure) {
            envHttpsProxy = proxyURI;
        } else {
            envHttpProxy = proxyURI;
        }
        return this;
    }

    public URLConnection openConnection(URI relative, Configure configCallback) throws IOException {
        if (relative != null) {
            try {
                return openConnectionWithProxies(urlBase.toURI().resolve(relative).toURL(), configCallback);
            } catch (URISyntaxException ex) {
                throw new IOException(ex);
            }
        } else {
            return openConnectionWithProxies(urlBase, configCallback);
        }
    }

    private class ConnectionContext {
        private final Configure configCallback;
        private final CountDownLatch countDown;
        private final URL url;
        private final List<Connector> tryConnectors = new ArrayList<>();

        // @GuardedBy(this)
        private Connector winner;
        // @GuardedBy(this)
        private URLConnection openedConnection;
        // @GuardedBy(this)
        private IOException exProxy;
        // @GuardedBy(this)
        private IOException exDirect;
        // @GuardedBy(this);
        private int outcomes;

        ConnectionContext(URL url, Configure configCallback, CountDownLatch latch) {
            this.configCallback = configCallback;
            this.countDown = latch;
            this.url = url;
        }

        synchronized URLConnection getConnection() throws IOException {
            if (openedConnection == null) {
                if (exDirect != null) {
                    throw exDirect;
                } else if (exProxy != null) {
                    throw exProxy;
                }
                throw new ConnectException(feedback.l10n("EXC_CannotConnectTo", url));
            } else {
                return openedConnection;
            }
        }

        boolean setOutcome(Connector w, URLConnection opened) {
            synchronized (this) {
                if (winner != null) {
                    return false;
                }
                winner = w;
                openedConnection = opened;
            }
            if (countDown != null) {
                countDown.countDown();
            }
            return true;
        }

        void setOutcome(boolean direct, IOException e) {
            synchronized (this) {
                if (direct) {
                    exDirect = e;
                } else {
                    exProxy = e;
                }
                if (++outcomes == tryConnectors.size()) {
                    countDown.countDown();
                }
            }
        }

        synchronized IOException getConnectException() {
            if (exDirect != null) {
                return exDirect;
            } else if (exProxy != null) {
                return exProxy;
            }
            return new ConnectException(feedback.l10n("EXC_TimeoutConnectTo", url));
        }

        synchronized void submit(Connector c) {
            tryConnectors.add(c);
        }

        void start() {
            for (Connector c : tryConnectors) {
                connectors.submit(c);
            }
        }
    }

    /**
     * Regexp to detect a scheme-like part. See https://tools.ietf.org/html/rfc3986#section-3.1.
     */
    private static final Pattern SCHEME_REGEXP = Pattern.compile("^\\p{Alpha}[\\p{Alnum}+-.]*:", Pattern.CASE_INSENSITIVE); // NOI18N

    public static InetSocketAddress proxyAddress(String proxySpec) throws URISyntaxException {
        if (proxySpec == null) {
            return null;
        }
        String trimmed = proxySpec.trim();
        if ("".equals(trimmed)) {
            return null;
        }
        // the URI is created just to parse the proxy specification. It won't be
        // actually used as a whole to form URL or open connection
        URI uri;

        try {
            uri = new URI(trimmed);
        } catch (URISyntaxException ex) {
            // maybe the user did not include the http:// in the address spec.
            if (SCHEME_REGEXP.matcher(trimmed).find()) {
                // something that could be a scheme is already in, but still won't parse
                throw ex;
            }
            try {
                trimmed = PROXY_SCHEME_PREFIX + trimmed;
                uri = new URI(trimmed);
            } catch (URISyntaxException ex2) {
                // no luck, throw the original exception:
                throw ex;
            }
            if (uri.getHost() == null || uri.getPort() < 1) {
                throw ex;
            }
        }

        // if the user forgets the scheme (http://) the string is misparsed and hostname
        // becomes the scheme while the host part will be empty. Adding the "http://" in
        // front
        // fixes parsing at least for the host/port part.
        if (uri.getScheme() == null || uri.getHost() == null) {
            URI checkURI = null;
            try {
                checkURI = new URI(PROXY_SCHEME_PREFIX + trimmed);
                if (!(checkURI.getHost() == null || checkURI.getPort() < 1)) {
                    uri = checkURI;
                }
            } catch (URISyntaxException ex) {
                // better leave the specified value without the scheme.
            }
        }
        InetSocketAddress address = InetSocketAddress.createUnresolved(uri.getHost(), uri.getPort());
        return address;
    }

    final class Connector implements Runnable {
        private final InetSocketAddress proxyAddress;
        private URL url;
        private ConnectionContext context;

        Connector(InetSocketAddress address) {
            this.proxyAddress = address;
        }

        InetSocketAddress getProxyAddress() {
            return proxyAddress;
        }

        boolean isDirect() {
            return proxyAddress == null;
        }

        boolean accepts(URL u) {
            synchronized (this) {
                return Objects.equals(u.getAuthority(), url.getAuthority());
            }
        }

        Connector bind(ConnectionContext ctx) {
            synchronized (this) {
                context = ctx;
                url = ctx.url;
            }
            return this;
        }

        @Override
        public void run() {
            ConnectionContext ctx;

            synchronized (this) {
                ctx = context;
                context = null;
            }
            runWithContext(ctx);
        }

        void runWithContext(ConnectionContext ctx) {
            final Proxy proxy;
            if (isDirect()) {
                proxy = null;
            } else {
                proxy = new Proxy(Proxy.Type.HTTP, proxyAddress);
            }
            Configure configCallback = ctx.configCallback;
            boolean won = false;
            URLConnection test = null;
            try {
                test = proxy == null ? url.openConnection() : url.openConnection(proxy);
                if (configCallback != null) {
                    configCallback.accept(test);
                }
                test.connect();
                if (test instanceof HttpURLConnection) {
                    HttpURLConnection htest = (HttpURLConnection) test;
                    int rcode = htest.getResponseCode();
                    // using bad request 400 as a marker. All 4xx, 5xx codes should be handled this
                    // way to force
                    // the appropriate exception out.
                    if (rcode >= HttpURLConnection.HTTP_BAD_REQUEST) {
                        // force the exception, should fail with IOException
                        try {
                            htest.getInputStream().close();
                        } catch (IOException ex) {
                            throw new HttpConnectionException(ex.getMessage(), ex, isDirect(), htest);
                        }
                        if (!isDirect()) { // this should not happen
                            throw new IOException(feedback.l10n("EXC_ProxyFailed", rcode));
                        }
                    }
                }
                won = ctx.setOutcome(this, test);
            } catch (IOException ex) {
                ctx.setOutcome(isDirect(), ex);
            } finally {
                if (!won) {
                    if (test instanceof HttpURLConnection) {
                        ((HttpURLConnection) test).disconnect();
                    }
                }
            }
        }
    }

    public static class HttpConnectionException extends IOException {
        private static final long serialVersionUID = 1L;
        private final boolean isDirect;
        private final int retCode;
        private final URL connectionUrl;
        private final String response;

        public HttpConnectionException(String message, IOException cause, boolean isDirect, HttpURLConnection connection) throws IOException {
            super(message, cause);
            this.response = parseErrorResponse(connection);
            this.retCode = connection.getResponseCode();
            this.connectionUrl = connection.getURL();
            this.isDirect = isDirect;
        }

        private static String parseErrorResponse(HttpURLConnection connection) {
            try (BufferedReader br = new BufferedReader(new InputStreamReader(connection.getErrorStream()))) {
                return br.lines().reduce((s1, s2) -> s1 + "\n" + s2).orElse("");
            } catch (IOException t) {
                return t.getLocalizedMessage();
            }
        }

        public int getRetCode() {
            return retCode;
        }

        public String getResponse() {
            return response;
        }

        public URL getConnectionUrl() {
            return connectionUrl;
        }

        public boolean isProxy() {
            return isDirect;
        }
    }

    List<Connector> makeConnectors(String httpProxy, String httpsProxy) {
        InetSocketAddress httpProxyAddress = null;
        InetSocketAddress httpsProxyAddress = null;

        URISyntaxException httpError = null;
        URISyntaxException httpsError = null;

        if (httpProxy != null) {
            try {
                httpProxyAddress = proxyAddress(httpProxy);
            } catch (URISyntaxException ex) {
                httpError = ex;
            }
        }
        if (httpsProxy != null) {
            if (!Objects.equals(httpProxy, httpsProxy)) {
                try {
                    httpsProxyAddress = proxyAddress(httpsProxy);
                } catch (URISyntaxException ex) {
                    httpsError = ex;
                }
            }
        }
        List<Connector> tryConnectors = new ArrayList<>();

        if (httpProxyAddress != null) {
            tryConnectors.add(new Connector(httpProxyAddress));
        }
        // do not attempt 2nd probe try if http+https are set to the same value.
        if (httpsProxyAddress != null) {
            tryConnectors.add(new Connector(httpsProxyAddress));
        }

        // something is there, but an error happened
        if (httpError != null) {
            feedback.error("WARN_HttpProxyGarbage", httpError, httpProxy);
        }
        if (httpsError != null) {
            feedback.error("WARN_HttpsProxyGarbage", httpError, httpProxy);
        }
        if (tryConnectors.isEmpty()) {
            // let the native support do its magic:
            System.setProperty("java.net.useSystemProxies", "true");
        }
        tryConnectors.add(new Connector(null));
        return tryConnectors;
    }

    private URLConnection openConnectionWithProxies(URL url, Configure configCallback) throws IOException {
        final CountDownLatch connected = new CountDownLatch(1);
        String httpProxy;
        String httpsProxy;

        Connector winner;
        ConnectionContext ctx = new ConnectionContext(url, configCallback, connected);

        synchronized (this) {
            httpProxy = envHttpProxy;
            httpsProxy = envHttpsProxy;
            winner = winningConnector;
        }

        boolean haveProxy = false;

        // reuse the same detected direct / proxy for matching URLs.
        if (winner != null && winner.accepts(url)) {
            winner.bind(ctx);
            // submit the winner so we can also recover from long connect delays
            ctx.submit(winner);
            // note: the winner will benefit from larger timeout as it is just one
            // connection
        } else {
            List<Connector> newConnectors = makeConnectors(httpProxy, httpsProxy);
            haveProxy = newConnectors.size() > 1;
            newConnectors.forEach(c -> ctx.submit(c.bind(ctx)));
        }

        ctx.start();

        int shouldDelay = haveProxy ? connectDelay : directConnectDelay;

        URLConnection res = null;

        try {
            if (shouldDelay > 0) {
                if (!connected.await(shouldDelay, TimeUnit.SECONDS)) {
                    throw ctx.getConnectException();
                } else {
                    // may also throw exception
                    res = ctx.getConnection();
                }
            } else {
                // wait indefinitely ... until network times out.
                connected.await();
                res = ctx.getConnection();
            }
        } catch (InterruptedException ex) {
            throw new ConnectException(feedback.l10n("EXC_InterruptedConnectingTo", url));
        }
        return res;
    }

    @Override
    public URLConnection createConnection(URL u, Configure configCallback) throws IOException {
        try {
            return openConnection(u.toURI(), configCallback);
        } catch (URISyntaxException ex) {
            throw new IOException(ex);
        }
    }

}
